using System;
using System.Collections;
using System.Runtime.InteropServices;
using System.Text;
using System.Xml;

namespace ReadSvcXml
{
    public class SetMsiData : IDisposable
    {
        // #define MSIDBOPEN_TRANSACT     (LPCTSTR)1 
        [DllImport("msi")]
        private static extern int MsiOpenDatabase(string dbpath, int persist, ref IntPtr msihandle);
        [DllImport("msi")]
        private static extern int MsiDatabaseCommit(IntPtr handle);
        [DllImport("msi")]
        private static extern int MsiDatabaseOpenView(IntPtr handle, string query, ref IntPtr viewhandle);
        [DllImport("msi")]
        private static extern int MsiViewExecute(IntPtr viewhandle, IntPtr recordhandle);
        [DllImport("msi")]
        private static extern int MsiViewFetch(IntPtr viewhandle, ref IntPtr recordhandle);
        [DllImport("msi", CharSet = CharSet.Auto)]
        private static extern int MsiRecordGetString(IntPtr recordhandle, int recno, StringBuilder szbuff, ref int len);
        [DllImport("msi")]
        private static extern int MsiCloseHandle(IntPtr handle);
        [DllImport("msi")]
        private static extern int MsiViewClose(IntPtr viewhandle); //Not used
        [DllImport("msi")]
        private static extern IntPtr MsiGetLastErrorRecord();
        [DllImport("msi")]
        private static extern int MsiFormatRecord(IntPtr empty, IntPtr errhandle, StringBuilder extradata, ref int len);

        string thepath = null;
        IntPtr dbhandle = IntPtr.Zero;

        public SetMsiData(string package)
        {
            thepath = package;
            int nres = MsiOpenDatabase(thepath, 1, ref dbhandle);
        }
        ~SetMsiData()
        {
            InternalDispose(false);
        }
        public string FindServiceNameComponent(string filename)
        {
            string query = "SELECT `Component_`, `FileName` from `File`";
            int rel;
            IntPtr viewhandle = IntPtr.Zero;
            IntPtr nothing = IntPtr.Zero;
            int nres = MsiDatabaseOpenView(dbhandle, query, ref viewhandle);
            nres = MsiViewExecute(viewhandle, nothing);
            IntPtr rechandle = IntPtr.Zero;
 
            bool GotSvc = false;
            string ComponentName=null;
            do
            {
                nres = MsiViewFetch(viewhandle, ref rechandle);
                if (0 != nres) // No data
                {
                    rel = MsiCloseHandle(viewhandle);
                    return null;
                }
                int outlen = 512;
                StringBuilder cmpnt = new StringBuilder(outlen);
                // Get the component name
                nres = MsiRecordGetString(rechandle, 1, cmpnt, ref outlen);
                outlen = 512;
                StringBuilder fn = new StringBuilder(512);
                //Get the filename 
                nres = MsiRecordGetString(rechandle, 2, fn, ref outlen);
                string[] thispair = fn.ToString().Split('|');
                string p1 = thispair[0].Trim();
                string p2 = thispair[1].Trim();
                if (p2 != null)
                {
                    if (p2.ToLower()==filename.ToLower())
                    {
                        GotSvc=true;
                        ComponentName = cmpnt.ToString();
                    }
                    if (p1.ToLower()==filename.ToLower())
                    {
                        GotSvc=true;
                        ComponentName = cmpnt.ToString();
                    }
                }
            }
            while (!GotSvc); 

            rel = MsiCloseHandle(viewhandle);
            rel = MsiCloseHandle(rechandle);
            rel = MsiCloseHandle(nothing);
            return ComponentName;

        }
        public string LastMsiErrorText()
        {
            IntPtr errhandle = IntPtr.Zero;
            int len = 1024;
            StringBuilder errmsg = new StringBuilder(len);
            try
            {
                errhandle = MsiGetLastErrorRecord();
                int resf = MsiFormatRecord(IntPtr.Zero, errhandle, errmsg, ref len);
            }
            finally
            {
                MsiCloseHandle(errhandle);
            }
            return errmsg.ToString();
        }

        public bool DoUpdate(string query)
        {
         //   MessageBox.Show (query);
            int rel = 0;
            IntPtr viewhandle = IntPtr.Zero;
            IntPtr nothing = IntPtr.Zero;
            try
            {
                int nres = MsiDatabaseOpenView(dbhandle, query, ref viewhandle);
                if (0 != nres)
                {
                    string err = LastMsiErrorText();
                    throw new ApplicationException("Update Failed " + err);
                } 
                nres = MsiViewExecute(viewhandle, nothing);
                if (0 != nres)
                {
                    string err = LastMsiErrorText();
                    throw new ApplicationException("Update Failed " + err); 
                }
            }
            finally
            {
                // Close handles used by this method
                rel = MsiCloseHandle(viewhandle);
                rel = MsiCloseHandle(nothing);
            }
            return true;

        }
        public void InternalDispose(bool disposing)
        {
            int res = MsiDatabaseCommit(dbhandle);
            res = MsiCloseHandle(dbhandle);
            dbhandle = IntPtr.Zero;
        }
        public void Dispose()
        {
            InternalDispose(true);
            GC.SuppressFinalize(this);
        }
    }// class SetMsiData
    public class GetXml
    {
        System.Xml.XmlTextReader rdr = null;

        public GetXml(string filename)
        {
            rdr = new XmlTextReader(filename);
        }
        private string NextVal()
        {
            do
            {
                rdr.Read();
            }
            while (rdr.NodeType != System.Xml.XmlNodeType.Text);
            return rdr.Value.Trim();
        }
        private bool NextBool(string thing)
        {
            rdr.Read();
            string b = rdr.Value.ToLower().Trim();

            if ("yes" == b)
                return true;
            if ("no"==b)
                return false;
            throw new ApplicationException (thing + " must be yes or no");

        }
        private ServiceInstallSpec ParseInstall()
        {
            bool BreakOut = false;
            ServiceInstallSpec val = new ServiceInstallSpec();
            // When we enter, we have just parsed "ServiceInstall" 
            while (rdr.Read())
            {
                string nodename = rdr.Name.ToLower();
                string nodevalue = rdr.Value;
                switch (rdr.NodeType)
                {
                    case System.Xml.XmlNodeType.Element:
                        string elem = nodename;
                        switch (elem)
                        {
                            case "id":
                                val.Id = NextVal();
                            break;
                            case "name":
                                val.Name = NextVal();
                            break;
                            case "displayname":
                                val.DisplayName = NextVal();
                            break;
                            case "servicetype":
                                val.ServiceType = NextVal();
                            break;
                            case "interactive":
                                val.Interactive = NextBool (elem);
                            break;
                            case "startname":
                                val.Startname = NextVal();
                            break;
                            case "start":
                                val.Start= NextVal();
                            break;
                            case "errorcontrol":
                                val.ErrorControl = NextVal();
                            break;
                            case "loadordergroup":
                                val.LoadOrderGroup = NextVal();
                            break;
                            case "dependencies":
                                val.Dependencies [0] = NextVal();
                            break;
                            case "password":
                                val.Password = NextVal();
                            break;
                            case "description":
                                val.Description = NextVal();
                                break;
                            case "arguments":
                                val.Arguments = NextVal();
                            break;
                            default:
                                throw new ApplicationException("Unknown Service attribute " + elem);
                        }
                        break;
                    case System.Xml.XmlNodeType.Text:
                        break;
                    case System.Xml.XmlNodeType.EndElement:
                        if (nodename == "serviceinstall")
                            BreakOut = true;
                        break;
                    default:
                        break;
                }
                if (BreakOut)
                    break; // the while loop
            }
            return val;
        }
        private string NextControlChoice(string thing)
        {
            rdr.Read();
            string b = rdr.Value.ToLower().Trim();

            if ("both" == b)
                return b;
            if ("install" == b)
                return b;
            if ("uninstall" == b)
                return b;
            throw new ApplicationException(thing + " must be Install, Uninstall or Both");
        }
        private ServiceControlSpec ParseControl()
        {
            bool BreakOut = false;
            ServiceControlSpec val = new ServiceControlSpec();
            // When we enter, we have just parsed "ServiceInstall" 
            while (rdr.Read())
            {
                string nodename = rdr.Name.ToLower();
                string nodevalue = rdr.Value;
                switch (rdr.NodeType)
                {
                    case System.Xml.XmlNodeType.Element:
                        string elem = nodename;
                        switch (elem)
                        {
                            case "id":
                                val.Id = NextVal();
                                break;
                            case "start":
                                val.Start = NextControlChoice(elem);
                                break;
                            case "stop":
                                val.Stop = NextControlChoice(elem);
                                break;
                            case "name":
                                val.Name = NextVal();
                                break;
                            case "remove":
                                val.Remove = NextControlChoice(elem);
                                break;
                            case "arguments":
                                val.Arguments = NextVal();
                                break;
                            case "wait":
                                val.Wait = NextBool(elem);
                                break;
                            default:
                                throw new ApplicationException("Unknown ServiceControl attribute " + elem);
                        }
                        break;
                    case System.Xml.XmlNodeType.Text:
                        break;
                    case System.Xml.XmlNodeType.EndElement:
                        if ("servicecontrol"==nodename)
                            BreakOut = true;
                        break;
                    default:
                        break;
                }
                if (BreakOut)
                    break;
            }
            return val;
        }
        public ServiceContainer Parse()
        {
            ServiceControlSpec sc = null;
            ServiceInstallSpec si = null;
            string svcfilename = null;
            bool BreakOut = false;
            while (rdr.Read())
            {
                string thename = rdr.Name.ToLower();
                switch (rdr.NodeType)
                {
                    case System.Xml.XmlNodeType.Element:
                        if (thename == "serviceinstall")
                        {
                            si = ParseInstall();
                        }
                        else
                        if (thename == "servicecontrol")
                        {
                            sc = ParseControl();
                        }
                        else
                        if (thename == "servicedata")
                        {
                            try
                            {
                                do
                                {
                                    rdr.Read();
                                }
                                while (rdr.NodeType != System.Xml.XmlNodeType.Element & rdr.Name.ToLower() != "filename");
                            }
                            catch (Exception e)
                            {
                                throw new ApplicationException("Filename not specified in ServiceData", e);
                            }
                            svcfilename = NextVal();
                        }
                        else
                            throw new ApplicationException("ServiceControl or ServiceInstall required");
                        break;
                    case XmlNodeType.Document:
                        break;
                    case XmlNodeType.DocumentType:
                        break;
                    case XmlNodeType.EntityReference:
                        break;
                    case System.Xml.XmlNodeType.Text:
                        break;
                    case System.Xml.XmlNodeType.Comment:
                        break;
                    case System.Xml.XmlNodeType.EndElement:
                        if (thename == "servicedata")
                            BreakOut = true;
                        break;
                    case XmlNodeType.XmlDeclaration:
                        break;
                    case XmlNodeType.CDATA:
                        break;
                    default:
                        break;
                }
                if (BreakOut)
                    break;
            }
            ServiceContainer scon = new ServiceContainer(si, sc, svcfilename);
            scon.Validate();
            return scon;
        }

    }
    public class ServiceInstallSpec
    {
        // Really dumb container for Service install data 
        public string Id = null;
        public string Name=null;
        public string DisplayName = null;
        public string ServiceType = "ownprocess";
        public long ServiceTypeLong = 0; 
        public bool Interactive = false;
        public string ErrorControl = "normal";
        public long ErrorControlLong = 0; 
        public string[] Dependencies = new string[2];
        public string LoadOrderGroup = null;
        public string Start = "auto";
        public long StartLong = 0;
        public string Startname = null;
        public string Password = null;
        public string Arguments = null;
        public string Description = null; 
        public void Validate()
        {
            if (null == Name)
                throw new ApplicationException("Name is required in ServiceInstall");
            if (null == Id)
                throw new ApplicationException("Id is required in ServiceInstall");
            if (ServiceType == "ownprocess")
                ServiceTypeLong = 0x00000010;
            else
            if (ServiceType == "shareprocess")
                ServiceTypeLong = 0x00000020;
            else
               throw new ApplicationException("ServiceType must be OwnProcess or ShareProcess");
            if (Interactive)
                ServiceTypeLong |= 0x00000100;
            if (ErrorControl == "normal")
                ErrorControlLong = 1; 
            else
                if (ErrorControl == "ignore")
                    ErrorControlLong = 0;
                else
                    if (ErrorControl == "critical")
                        ErrorControlLong = 3;
                    else
                        throw new ApplicationException("ErrorControl must be Normal, ignore or critical");
            if (Password!=null && Startname==null)
                throw new ApplicationException("Password is set but not StartName");
            if (Start == "auto")
                StartLong = 2;
            else
                if (Start == "demand")
                    StartLong = 3;
                else
                    if (Start == "disabled")
                        StartLong = 4;        
            else 
                        throw new ApplicationException("Start must be Auto, Demand or Disabled");
        }
    }
    public class ServiceControlSpec
    {
        // MSI Event values 
        const int InstallStart = 1;
        const int InstallStop = 2;
        const int InstallDelete = 8; 
        const int UninstallStart = 16;
        const int UninstallStop = 32;
        const int UninstallDelete = 128; 

        public string Id = null;
        public string Name = null;
        public string Arguments = null;
        public string Start = null;
        public string Stop = null;
        public string Remove = null;
        public long RemoveLong = 0;
        public bool Wait = false;
        public long WaitLong = 0;
        public long EventLong = 0; 


        public void Validate()
        {
            if (Name==null)
                throw new ApplicationException("ServiceControl must specify Name for the Service ");
            if (Id==null)
                throw new ApplicationException("ServiceControl must specify an Id for the Service ");
            if (Wait)
                WaitLong = 1;
            if (Start == "install")
                EventLong += InstallStart;
            else
                if (Start == "uninstall")
                    EventLong += UninstallStart;
                else
                    if (Start == "both")
                        EventLong += (InstallStart + UninstallStart);
                    else // Shouldn't really get here
                        throw new ApplicationException("Start Must be Install, Uninstall or Both");
            if (Stop == "install")
                EventLong += InstallStop;
            else
                if (Stop == "uninstall")
                    EventLong += UninstallStop;
                else
                    if (Stop == "both")
                        EventLong += (InstallStop + UninstallStop);
                    else // Shouldn't really get here
                        throw new ApplicationException("Stop Must be Install, Uninstall or Both");
            if (Remove==null)
                RemoveLong = UninstallDelete;
            else 
            {
                if (Remove == "install")
                    RemoveLong += InstallDelete;
                else
                    if (Remove == "both")
                        RemoveLong += (InstallDelete + UninstallDelete);
                    else
                        if (Remove == "uninstall")
                            RemoveLong += UninstallDelete;
                        else
                            throw new ApplicationException("Remove must be Install, Unistall or Both");
            }
             EventLong += RemoveLong;
         }
    }
    public class ServiceContainer
    {
        public ServiceContainer(ServiceInstallSpec s, ServiceControlSpec c, string fname)
        {
            sis = s;
            scs = c;
            ServiceFilename = fname;
        }
        public ServiceInstallSpec sis = null;
        public ServiceControlSpec scs = null;
        public string ServiceFilename = null;
        public void Validate()
        {
            if (null == ServiceFilename)
                throw new ApplicationException("FileName is required in ServiceData");
            sis.Validate();
            scs.Validate();
        }
    }
    class Program
    {
        static void AddString(StringBuilder st, string val)
        {
            st.Append("," + "'" + val + "' ");
        }
        static void AddBool(StringBuilder st, bool b)
        {
            st.Append("," + "'"  + b.ToString() + "' ");
        }
        static void AddNumber(StringBuilder st, long num)
        {
            st.Append(", "+ num.ToString() + " ");
        }
        static void Main(string[] args)
        {
            // First arg the name of the MSI file 
            //  Second arg the name of the XML file
            GetXml gx = new GetXml(args [1] );
             ServiceContainer myspecs = gx.Parse();
            // If we get here, whatever specs we have are valid
            SetMsiData sm = new SetMsiData(args[0]);
            ServiceInstallSpec si = myspecs.sis;
            string MsiComponent = sm.FindServiceNameComponent(myspecs.ServiceFilename);
            if (null == MsiComponent)
            {
                throw new ApplicationException("Service file name not found in MSI File Table");
            }
            StringBuilder insert = new StringBuilder();
            string tryit = "Insert into `ServiceInstall` (`ServiceInstall`,`Name`, `DisplayName`, `ServiceType`, `StartType`, `ErrorControl`, `LoadOrderGroup`, `Dependencies`, `StartName`, `Password`, `Arguments`, `Component_`, `Description` )";
            insert.Append(tryit);
            insert.Append(" VALUES ("); insert.Append("'" + si.Id + "'");
            AddString(insert, si.Name);
            AddString(insert, si.DisplayName);
            AddNumber(insert, si.ServiceTypeLong); AddNumber(insert, si.StartLong);
            AddNumber(insert, si.ErrorControlLong); AddString(insert, si.LoadOrderGroup);
            if (si.Dependencies[0] != null)
            {
                AddString(insert, si.Dependencies[0] + "[~][~]"); // Ends with this double null 
            }
            else
                AddString(insert, ""); 

            AddString(insert, si.Startname); AddString(insert, si.Password); AddString(insert, si.Arguments);
            AddString(insert, MsiComponent);
            AddString(insert, si.Description);
            insert.Append(")"); 
            sm.DoUpdate (insert.ToString());

            ServiceControlSpec sc = myspecs.scs;
            StringBuilder cntl = new StringBuilder();
            string svc = "Insert into `ServiceControl` (`ServiceControl`, `Name`, `Event`, `Arguments`, `Wait`, `Component_`)";
            cntl.Append(svc);
            cntl.Append(" VALUES(" + "'" + sc.Id + "'");
            AddString(cntl, sc.Name); AddNumber(cntl, sc.EventLong); AddString (cntl, sc.Arguments);
            AddNumber (cntl, sc.WaitLong); AddString (cntl, MsiComponent);
            cntl.Append (")");
            sm.DoUpdate (cntl.ToString());
            sm.Dispose();
            }

        }
    }
